/*
 * Dirty-hacks-collection tools.
 * Reparses glibc's charmap data.
 * Copyleft 2005 Vladimir A. Smirnov.
 * You may do whatever you want with this hackware.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static char *copy_token(char *dest,char *src,char endch)
{
	char *saved_src = src;
	while(*src <= 32) /* skip leading whitespace */
		src++;
	while(*src != endch)
	{
		if(!*src)
		{
			fprintf(stderr,"copy_token(): unexpected EOL while parsing \"%s\".\n",saved_src);
			abort();
		}
		*(dest++) = *(src++);
	}
	*dest = 0;
	return src;
}

static void usage(char *progname)
{
	puts("\n  This program reads glibc's locale description (from localedata/charmaps),");
	puts("  searches for strings starting with \"<U\" and than stores charmap's");
	puts("  data in form suitable for further inclusion in C program.\n");
	printf("  Usage: %s [--help] [--with-header] [input_file] [output_file]\n",progname);
	puts("\n  Pass '--with-header' to print necessary structure definition.\n");
}

int main(int argc,char *argv[])
{
	int i, len, print_header = 0;
	char buffer[1024], *ptr;
	char unibuf[64], abuf[64], namebuf[256];
	FILE *in = stdin, *out = stdout;
	for(i = 1;i < argc;i++)
	{
		if(argv[i][0] != '-' || argv[i][1] != '-') /* non-option */
			continue;
		if(!strcmp(argv[i],"--help")) {
			usage(argv[0]);
			return 0;
		} else if(!strcmp(argv[i],"--with-header")) {
			print_header = 1;
		} else {
			fprintf(stderr,"Unknown option \"%s\"\n",argv[i]);
			return -1;
		}
	}
	for(i = 1;i < argc;i++)
	{
		if(argv[i][0] == '-' && argv[i][1] == '-') /* option */
			continue;
		if(in == stdin)
		{
			/* assume input file name */
			in = fopen(argv[i],"rt");
			if(!in)
			{
				perror("Unable to open input file");
				return -1;
			}
			continue;
		}
		if(out == stdout)
		{
			/* assume output file name */
			out = fopen(argv[i],"wt");
			if(!out)
			{
				perror("Unable to open output file");
				return -1;
			}
			continue;
		}
		fprintf(stderr,"Too many arguments\n");
		break;
	}
	if(print_header)
	{
		fputs("typedef struct charmap_char_description_s\n",out);
		fputs("{\n\tunsigned char ascii_code;\n",out);
		fputs("\tunsigned short uni_code;\n",out);
		fputs("\tchar *char_description;\n",out);
		fputs("} charmap_char_description_t, *charmap_char_description_p;\n\n",out);
	}
	fputs("static charmap_char_description_t _ccd[] = \n{\n",out);
	while(!feof(in))
	{
		memset(buffer,0,1024);
		/* reserve space for copy_token()'s sentinel */
		fgets(buffer,1023,in);
		len = strlen(buffer);
		ptr = buffer + len;
		while(ptr > buffer && *ptr <= 32)
			*(ptr--) = 0;
		if(strlen(buffer) <= 4 || buffer[0] != '<' || buffer[1] != 'U')
			continue;
		ptr = copy_token(unibuf,buffer,'>');
		/* BUGBUG: just interpret 2nd token as _ascii_ code */
		ptr = copy_token(abuf,ptr + 1,' ');
		ptr = copy_token(namebuf,ptr,0);
		fprintf(out,"\t{0%s, 0x%s, \"%s\"},\n",abuf + 1,unibuf + 2,namebuf);
	}
	fputs("\t{0, 0, NULL}\n};\n",out);
	fclose(in);
	fclose(out);
	return 0;
}
